
class LoaderResult:

    ERROR_NOT_FOUND = "404 (not found): the image requested does not exist"
    ERROR_UPSTREAM = "upstream"
    ERROR_TIMEOUT = "timeout"

    def __init__(self, buffer=None, successful=True, error=None, metadata=None):
        """
        :param buffer: The media buffer

        :param successful: True when the media has been loaded.
        :type successful: bool

        :param error: Error code
        :type error: str

        :param metadata: Dictionary of metadata about the buffer
        :type metadata: dict
        """

        if metadata is None:
            metadata = {}

        self.buffer = buffer
        self.successful = successful
        self.error = error
        self.metadata = metadata




class Error(Exception):
    '''Base exception class'''
    pass

class BadRequestError(Error):
    '''The syntax was not respected'''
    pass

class UnknownUserError(Error):
    '''The user does not exist'''
    pass

class PaymentError(Error):
    '''Not enough funds on identity'''
    pass

class ForbiddenSignatureError(Error):
    '''The signature is incorrect'''
    pass

class NotFoundError(Error):
    '''The image requested does not exist'''
    pass

class MethodNotAllowedError(Error):
    '''A method other than GET or HEAD was sent for the resource'''
    pass

class UnsupportedMediaTypeError(Error):
    '''The media type specified in the request is not supported'''
    pass

class TooManyRequestsError(Error):
    '''Client is temporarily banned'''
    pass

class DashPlatformError(Error):
    ''' An error occured with the dash platform service'''
    pass

class UnsignedURLError(Error):
    '''URL is not signed'''
    pass

class UnsafeURLError(Error):
    '''URL has unsafe but unsafe is not allowed '''
    pass

class UnspecifiedImageError(Error):
    '''No original image was specified'''
    pass

class BlacklistedSourceError(Error):
    '''Source image url has been blacklisted'''
    pass



from dapiclient.client import DAPIClient
import cbor2
from thumbor_dash.error_handlers.sentry import ErrorHandler
from thumbor_dash.error_handlers import *

client = DAPIClient()


def getDocuments(handler, data):
    error_handler = ErrorHandler(handler.context.config)

    try:
        identity = client.getIdentity(data['owner_id'])

    except Exception as e:
        error_handler.handle_error(handler.context, handler, UnknownUserError)
        return

    else:
        docs = client.getDocuments(
            data['contract_id'],
            data['document_type'],
            data['where'],
            limit=2, # Only one document
            )

        documents = []
    
        for  doc in docs:
           documents.append(cbor2.loads(doc))
    
        return documents[0] # Return the only document in the list







from thumbor.handlers import ImageApiHandler
from thumbor.engines import BaseEngine
from thumbor_dash.error_handlers import *
from thumbor_dash.error_handlers.sentry import ErrorHandler

##
# Base handler for Image API operations
##
class ThumborDashImageApiHandler(ImageApiHandler):

    def validate(self, body):
        error_handler = ErrorHandler(self.context.config)

        conf = self.context.config
        mime = BaseEngine.get_mimetype(body)

        if mime == "image/gif" and self.context.config.USE_GIFSICLE_ENGINE:
            engine = self.context.modules.gif_engine
        else:
            engine = self.context.modules.engine

        # Check if image is valid
        try:
            engine.load(body, None)
        except IOError:
            error_handler.handle_error(self.context, self, UnsupportedMediaTypeError)
            return False

        # Check weight constraints
        if conf.UPLOAD_MAX_SIZE != 0 and len(self.request.body) > conf.UPLOAD_MAX_SIZE:
            self._error(
                412,
                "Image exceed max weight (Expected : %s, Actual : %s)"
                % (conf.UPLOAD_MAX_SIZE, len(self.request.body)),
            )
            return False

        # Check size constraints
        size = engine.size
        if conf.MIN_WIDTH > size[0] or conf.MIN_HEIGHT > size[1]:
            self._error(
                412,
                "Image is too small (Expected: %s/%s , Actual : %s/%s) % "
                "(conf.MIN_WIDTH, conf.MIN_HEIGHT, size[0], size[1])",
            )
            return False
        return True


import datetime

from thumbor.engines import BaseEngine
from thumbor_dash.handlers import ThumborDashImageApiHandler
from thumbor_dash.error_handlers import *
from thumbor_dash.error_handlers.sentry import ErrorHandler

##
# Handler to retrieve or modify existing images
# This handler support GET, PUT and DELETE method to manipulate existing images
##
class ThumborDashImageResourceHandler(ThumborDashImageApiHandler):
    async def check_resource(self, file_id):
        error_handler = ErrorHandler(self.context.config)

        file_id = file_id[: self.context.config.MAX_ID_LENGTH]
        # Check if image exists
        exists = await self.context.modules.storage.exists(file_id)

        if exists:
            body = await self.context.modules.storage.get(file_id)
            self.set_status(200)

            mime = BaseEngine.get_mimetype(body)
            if mime:
                self.set_header("Content-Type", mime)

            max_age = self.context.config.MAX_AGE
            if max_age:
                self.set_header("Cache-Control", "max-age=" + str(max_age) + ",public")
                self.set_header(
                    "Expires",
                    datetime.datetime.utcnow() + datetime.timedelta(seconds=max_age),
                )
            self.write(body)
            self.finish()
        else:
            error_handler.handle_error(self.context, self, NotFoundError)

    async def put(self, file_id):
        error_handler = ErrorHandler(self.context.config)

        file_id = file_id[: self.context.config.MAX_ID_LENGTH]
        # Check if image overwriting is allowed
        if not self.context.config.UPLOAD_PUT_ALLOWED:
            error_handler.handle_error(self.context, self, MethodNotAllowedError)
            return

        # Check if the image uploaded is valid
        if self.validate(self.request.body):
            await self.write_file(file_id, self.request.body)
            self.set_status(204)

    async def delete(self, file_id):
        error_handler = ErrorHandler(self.context.config)

        file_id = file_id[: self.context.config.MAX_ID_LENGTH]
        # Check if image deleting is allowed
        if not self.context.config.UPLOAD_DELETE_ALLOWED:
            error_handler.handle_error(self.context, self, MethodNotAllowedError)
            return

        # Check if image exists
        exists = await self.context.modules.storage.exists(file_id)
        if exists:
            await self.context.modules.storage.remove(file_id)
            self.set_status(204)
        else:
            error_handler.handle_error(self.context, self, NotFoundError)

    async def get(self, file_id):
        await self.check_resource(file_id)

    async def head(self, file_id):
        await self.check_resource(file_id)


from urllib.parse import quote, unquote
from thumbor_dash.error_handlers import *
from thumbor_dash.context import ThumborDashRequestParameters
from thumbor_dash.verifiers import access_status_verifier, url_field_verifier, image_size_verifier, thumbnail_size_verifier
from thumbor_dash import dapiclient 
from thumbor_dash.utils import dashauthParametersToJson
from thumbor.handlers.imaging import ImagingHandler
from thumbor_dash.error_handlers.sentry import ErrorHandler

import base58
import cbor2


class ThumborDashImagingHandler(ImagingHandler):

    async def check_image(self, kwargs):
        error_handler = ErrorHandler(self.context.config)

        if self.context.config.MAX_ID_LENGTH > 0:
            # Check if an image with an uuid exists in storage
            exists = await self.context.modules.storage.exists(
                kwargs["image"][: self.context.config.MAX_ID_LENGTH]
            )
            if exists:
                kwargs["image"] = kwargs["image"][: self.context.config.MAX_ID_LENGTH]

        url = self.request.path

        kwargs["image"] = quote(kwargs["image"].encode("utf-8"))
        if not self.validate(kwargs["image"]):
            error_handler.handle_error(self.context, self, UnspecifiedImageError)
            return

        kwargs["request"] = self.request
        self.context.request = ThumborDashRequestParameters(**kwargs)

        has_none = not self.context.request.unsafe and not self.context.request.hash
        has_both = self.context.request.unsafe and self.context.request.hash

        if has_none or has_both:
            error_handler.handle_error(self.context, self, UnsignedURLError)
            return

        if self.context.request.unsafe and not self.context.config.ALLOW_UNSAFE_URL:
            error_handler.handle_error(self.context, self, UnsafeURLError)
            return

        if self.context.config.USE_BLACKLIST:
            blacklist = await self.get_blacklist_contents()
            if self.context.request.image_url in blacklist:
                error_handler.handle_error(self.context, self, BlacklistedSourceError)
                return

        url_signature = self.context.request.hash
        if url_signature:
            signer = self.context.modules.url_signer(self.context.server.security_key)

            try:
                quoted_hash = quote(self.context.request.hash)
            except KeyError:
                error_handler.handle_error(self.context, self, ForbiddenSignatureError)
                return

            url_to_validate = url.replace(
                "/%s/" % self.context.request.hash, ""
            ).replace("/%s/" % quoted_hash, "")
            
            valid = signer.validate(unquote(url_signature).encode(), url_to_validate)

            if not valid and self.context.config.STORES_CRYPTO_KEY_FOR_EACH_IMAGE:
                # Retrieves security key for this image if it has been seen before
                security_key = await self.context.modules.storage.get_crypto(
                    self.context.request.image_url
                )
                if security_key is not None:
                    signer = self.context.modules.url_signer(security_key)
                    valid = signer.validate(url_signature.encode(), url_to_validate)

            if not valid:
                error_handler.handle_error(self.context, self, ForbiddenSignatureError)
                return


        # <--------------------- Dash Platform Request Verification  ------------------------->
        
        request = self.context.request # HTTP request
        config = self.context.config # thumbor config
      
        body = dashauthParametersToJson(request.dashauth) 
        
        thumbnail_width = request.width # requested thumbnail width
        thumbnail_height = request.height # requested thumbnail height
        
        MIN_WIDTH = config.MIN_WIDTH  # minimum thumbnail width allowed
        MAX_WIDTH = config.MAX_WIDTH # maximum thumbnail width allowed
        MIN_HEIGHT = config.MIN_HEIGHT  # minimum thumbnail height allowed
        MAX_HEIGHT = config.MAX_HEIGHT # maximum thumbnail height allowed

        requesterId = body["requester"] # identity of whomever is making the request
        contractId = body["contract"] # the contract whose document holds the image URL
        documentType = body["document"] # the document whose instance holds the URL
        field = body["field"] # the field of the URL
        ownerId = body["owner"] # the owner of the document that is being requested
        updatedAt = body["updatedAt"] # the last time the document was updated

        # Verify user access status
        checkAccessStatus = await access_status_verifier.verifyUserAccessStatus(requesterId, config)

        if checkAccessStatus:
             # DAPI thumbnail document request input data
             data = {
                 'contract_id': base58.b58decode(contractId),
                 'owner_id': base58.b58decode(ownerId),
                 'document_type': documentType,
                 'where': cbor2.dumps([
                     ['ownerId', '==', base58.b58decode(ownerId)],
                     ['$updatedAt', '==', updatedAt],
                     ]),
                }  

             try:
                 # Query DAPI for thumbnail document data
                 thumbnail_document = dapiclient.getDocuments(self, data)
             except Exception as e:
                 error_handler.handle_error(self.context, self, DashPlatformError)
                 return
             else:
                 #Request verification
                 MIN_RESIZE_WIDTH = thumbnail_document["resizeValues"][0]
                 MIN_RESIZE_HEIGHT = thumbnail_document["resizeValues"][1]
                 MAX_RESIZE_WIDTH = thumbnail_document["resizeValues"][2]
                 MAX_RESIZE_HEIGHT = thumbnail_document["resizeValues"][3]
                
                 checkURLField = url_field_verifier.verifyURLField(thumbnail_document, field) # Verify url field
                 checkThumbnailSize = thumbnail_size_verifier.verifyThumbnailSize(thumbnail_width, thumbnail_height, MIN_RESIZE_WIDTH, MIN_RESIZE_HEIGHT, MAX_RESIZE_WIDTH, MAX_RESIZE_HEIGHT) # Verify requested thumbnail size
                 checkImageSize = image_size_verifier.verifyImageSize(thumbnail_width, thumbnail_height, MIN_WIDTH, MIN_HEIGHT, MAX_WIDTH, MAX_HEIGHT) # Verify image size
      
                 if (checkURLField and checkThumbnailSize and checkImageSize) == False:
                     error_handler.handle_error(self.context, self, BadRequestError)
                     return
        else:
            error_handler.handle_error(self.context, self, TooManyRequestsError)
            return
       
        return await self.execute_image_operations()        
        


from tornado.routing import _RuleList

BUILTIN_HANDLERS = [
    "thumbor.handler_lists.healthcheck",
    "thumbor_dash.handler_lists.upload",
    "thumbor.handler_lists.blacklist",
]

HandlerList = _RuleList


from datetime import datetime

def dashauthParametersToJson(params):
    temp = params.split(":")

    result = {}

    for p in temp:
        key_value_pair = p.split("(")
        key = key_value_pair[0]

        temp_value = key_value_pair[1]
        value = temp_value[0:- 1]
        real_value = int(value) if key == "updatedAt" else value

        result[key] = real_value

    return result


def datetimeToMillisecondsSinceEpoch(dt):
    epoch = datetime.utcfromtimestamp(0)
    return (dt - epoch).total_seconds() * 1000.0




def verifyThumbnailSize(thumbnail_width, thumbnail_height, MIN_RESIZE_WIDTH, MIN_RESIZE_HEIGHT, MAX_RESIZE_WIDTH, MAX_RESIZE_HEIGHT):
    ''' Checks whether the requested thumbnail size is within allowed bounds'''
    if thumbnail_width >= MIN_RESIZE_WIDTH and thumbnail_width <= MAX_RESIZE_WIDTH and thumbnail_height >= MIN_RESIZE_HEIGHT and thumbnail_height <= MAX_RESIZE_HEIGHT:
        return True
    else:
        return False 



import base64
import hashlib
import hmac

from six import text_type

from libthumbor.url_signers import BaseUrlSigner


class UrlSigner(BaseUrlSigner):
    """Validate urls and sign them using base64 hmac-sha256
    """

    def signature(self, url):
        return base64.urlsafe_b64encode(
            hmac.new(
                self.security_key, text_type(url).encode("utf-8"), hashlib.sha256
            ).digest()
        )



import datetime
import re
import socket
from typing import Pattern
from urllib.parse import quote, unquote, urlparse

import tornado.httpclient

from thumbor_dash.loaders import LoaderResult
from thumbor.utils import logger


try:
    import tornado.curl_httpclient  # pylint: disable=ungrouped-imports
except (ImportError, ValueError):
    logger.warning(
        "pycurl usage is advised. It could not be loaded properly. Verify install..."
    )


def encode_url(url):
    if url == unquote(url):
        return quote(url, safe="~@#$&()*!+=:;,.?/'")

    return url


def encode(string):
    return None if string is None else string.encode("ascii")


def quote_url(url):
    return encode_url(unquote(url))


def _normalize_url(url):
    url = quote_url(url)
    return url if url.startswith("http") else "http://%s" % url


def validate(context, url, normalize_url_func=_normalize_url):
    url = normalize_url_func(url)
    res = urlparse(url)

    if not res.hostname:
        return False

    if not context.config.ALLOWED_SOURCES:
        return True

    for pattern in context.config.ALLOWED_SOURCES:
        if isinstance(pattern, Pattern):  # pylint: disable=isinstance-second-argument-not-valid-type
            match = url
        else:
            pattern = "^%s$" % pattern
            match = res.hostname

        if re.match(pattern, match):
            return True

    return False


def return_contents(response, url, context, req_start=None):
    res = urlparse(url)
    if req_start:
        finish = datetime.datetime.now()
        context.metrics.timing(
            "original_image.fetch.{0}.{1}".format(
                response.code, res.netloc.replace(".", "_")
            ),
            (finish - req_start).total_seconds() * 1000,
        )
        context.metrics.incr(
            "original_image.fetch.{0}.{1}".format(response.code, res.netloc.replace(".", "_"))
        )

    result = LoaderResult()
    context.metrics.incr("original_image.status." + str(response.code))
    context.metrics.incr(
        "original_image.status.{0}.{1}".format(response.code, res.netloc.replace(".", "_"))
    )
    if response.error:
        result.successful = False
        if response.code == 599:
            # Return a Gateway Timeout status downstream if upstream times out
            result.error = LoaderResult.ERROR_TIMEOUT
        else:
            result.error = LoaderResult.ERROR_NOT_FOUND

        logger.warning(u"ERROR retrieving image %s: %s", url, str(response.error))

    elif response.body is None or len(response.body) == 0:
        result.successful = False
        result.error = LoaderResult.ERROR_UPSTREAM

        logger.warning(u"ERROR retrieving image %s: Empty response.", url)
    else:
        if response.time_info:
            for metric_name in response.time_info:
                context.metrics.timing(
                    "original_image.time_info." + metric_name,
                    response.time_info[metric_name] * 1000,
                )
            context.metrics.timing(
                "original_image.time_info.bytes_per_second",
                len(response.body) // response.time_info["total"],
            )
        result.buffer = response.body
        result.metadata.update(response.headers)
        context.metrics.incr("original_image.response_bytes", len(response.body))

    return result


async def load(
    context,
    url,
    normalize_url_func=_normalize_url,
    return_contents_fn=return_contents,
    encode_fn=encode,
):
    using_proxy = (
        context.config.HTTP_LOADER_PROXY_HOST and context.config.HTTP_LOADER_PROXY_PORT
    )
    if using_proxy or context.config.HTTP_LOADER_CURL_ASYNC_HTTP_CLIENT:
        http_client_implementation = "tornado.curl_httpclient.CurlAsyncHTTPClient"
        prepare_curl_callback = _get_prepare_curl_callback(context.config)
        exc_type = tornado.curl_httpclient.CurlError
    else:
        http_client_implementation = None  # default
        prepare_curl_callback = None
        exc_type = tornado.httpclient.HTTPClientError

    tornado.httpclient.AsyncHTTPClient.configure(
        http_client_implementation, max_clients=context.config.HTTP_LOADER_MAX_CLIENTS,
    )
    client = tornado.httpclient.AsyncHTTPClient()

    user_agent = None
    headers = {"Accept": "image/*;q=0.9,*/*;q=0.1"}
    if context.config.HTTP_LOADER_FORWARD_ALL_HEADERS:
        headers = context.request_handler.request.headers
    else:
        if context.config.HTTP_LOADER_FORWARD_USER_AGENT:
            if "User-Agent" in context.request_handler.request.headers:
                user_agent = context.request_handler.request.headers["User-Agent"]
        if context.config.HTTP_LOADER_FORWARD_HEADERS_WHITELIST:
            for header_key in context.config.HTTP_LOADER_FORWARD_HEADERS_WHITELIST:
                if header_key in context.request_handler.request.headers:
                    headers[header_key] = context.request_handler.request.headers[
                        header_key
                    ]

    if user_agent is None and "User-Agent" not in headers:
        user_agent = context.config.HTTP_LOADER_DEFAULT_USER_AGENT

    url = normalize_url_func(url)
    req = tornado.httpclient.HTTPRequest(
        url=url,
        headers=headers,
        connect_timeout=context.config.HTTP_LOADER_CONNECT_TIMEOUT,
        request_timeout=context.config.HTTP_LOADER_REQUEST_TIMEOUT,
        follow_redirects=context.config.HTTP_LOADER_FOLLOW_REDIRECTS,
        max_redirects=context.config.HTTP_LOADER_MAX_REDIRECTS,
        user_agent=user_agent,
        proxy_host=encode_fn(context.config.HTTP_LOADER_PROXY_HOST),
        proxy_port=context.config.HTTP_LOADER_PROXY_PORT,
        proxy_username=encode_fn(context.config.HTTP_LOADER_PROXY_USERNAME),
        proxy_password=encode_fn(context.config.HTTP_LOADER_PROXY_PASSWORD),
        ca_certs=encode_fn(context.config.HTTP_LOADER_CA_CERTS),
        client_key=encode_fn(context.config.HTTP_LOADER_CLIENT_KEY),
        client_cert=encode_fn(context.config.HTTP_LOADER_CLIENT_CERT),
        validate_cert=context.config.HTTP_LOADER_VALIDATE_CERTS,
        prepare_curl_callback=prepare_curl_callback,
    )

    start = datetime.datetime.now()
    try:
        response = await client.fetch(req, raise_error=True)
    except exc_type as err:
        response = tornado.httpclient.HTTPResponse(
            req, err.code, reason=err.message, start_time=start
        )
    except socket.gaierror as err:
        response = tornado.httpclient.HTTPResponse(
            req, 599, reason=str(err), start_time=start
        )

    return return_contents_fn(
        response=response, url=url, context=context, req_start=start,
    )


def _get_prepare_curl_callback(config):
    if (
        config.HTTP_LOADER_CURL_LOW_SPEED_TIME == 0
        or config.HTTP_LOADER_CURL_LOW_SPEED_LIMIT == 0
    ):
        return None

    class CurlOpts:
        def __init__(self, config):
            self.config = config

        def prepare_curl_callback(self, curl):
            curl.setopt(
                curl.LOW_SPEED_TIME, self.config.HTTP_LOADER_CURL_LOW_SPEED_TIME,
            )
            curl.setopt(
                curl.LOW_SPEED_LIMIT, self.config.HTTP_LOADER_CURL_LOW_SPEED_LIMIT,
            )

    return CurlOpts(config).prepare_curl_callback


from thumbor_dash.error_handlers import *


class ErrorHandler:
    '''Custom error handler for thumbor_dash'''

    def __init__(self, config):
        self.config = config
        self.BAN_DURATION = config.BAN_DURATION #Ban duration in minutes


    def handle_error(self, context, handler, exception):
        # context is thumbor_dash's context for the current request
        # handler is tornado's request handler for the current request
        # exception is the error that occurred

        # Custom error messages for thumbor_dash server
        self.ERROR_MESSAGES = {
        BadRequestError: "400 (bad request): the syntax was not respected",
        UnsignedURLError: "400 (unsigned request): URL does not have hash or unsafe, or has both",
        UnsafeURLError: "400 (unsafe request) URL has unsafe but unsafe is not allowed by the config",
        UnspecifiedImageError: "400 (unspecified image) No original image was specified in the given URL",
        BlacklistedSourceError: "400 (blacklisted source) Source image url has been blacklisted",
        UnknownUserError: "401 (unauthorized): the user does not exist",
        PaymentError: "402 (payment required): in phase 2 only -> not enough funds on identity",
        ForbiddenSignatureError: "403 (forbidden): the signature is incorrect",
        NotFoundError: "404 (not found): the image requested does not exist",
        MethodNotAllowedError: "405 (method not allowed): a method other than GET or HEAD was sent for the resource",
        UnsupportedMediaTypeError: "415 (unsupported media type): client should not request again",
        TooManyRequestsError: "429 (too many requests): client should stop making requests for a {} min period".format(self.BAN_DURATION),
        DashPlatformError: "503 (Dash Platform Service Error): Dash platform service is not available. Please retry later",
        }
        
        exception_code_switcher = {
        BadRequestError: 400,
        UnsignedURLError: 400,
        UnsafeURLError: 400,
        UnspecifiedImageError: 400,
        BlacklistedSourceError: 400,
        UnknownUserError: 401,
        PaymentError: 402,
        ForbiddenSignatureError: 403,
        NotFoundError: 404,
        MethodNotAllowedError: 405,
        UnsupportedMediaTypeError: 415,
        TooManyRequestsError: 429,
        DashPlatformError: 503
        }

        error_status_code = exception_code_switcher.get(exception, BadRequestError)

        handler.clear()
        handler.set_status(error_status_code)
        handler.finish("<html><body>{}</body></html>".format(self.ERROR_MESSAGES[exception])
        )




import tornado.ioloop
import tornado.web

from thumbor_dash.url import ThumborDashUrl
from thumbor_dash.handlers.imaging import ThumborDashImagingHandler
from thumbor.app import ThumborServiceApp


class ThumborDashServiceApp(ThumborServiceApp):

    def get_handlers(self):
        handlers = []
        for handler_list in self.context.modules.importer.handler_lists:
            get_handlers = getattr(handler_list, "get_handlers", None)
            if get_handlers is None:
                continue
            handlers.extend(get_handlers(self.context))

        # Imaging handler (GET)
        handlers.append((ThumborDashUrl.regex(), ThumborDashImagingHandler, {"context": self.context}))

        return handlers